#ifdef COMMENT_ONLY
* For FlagShip 5 and 6 
* Copyright (c) 1996..2007 by multisoft Datentechnik, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove this copyright notice.
* 
* Release $Date: 2007/09/04 11:13:26 $
*
#endif           
                                                                                            
*-----------------------------------------------------------------------------
* The following AUTORLOCK(), AUTOFLOCK() and AUTOUNLOCK() functions are
* invoked automatically from the (default) RDD driver DBFIDX in SHARED mode
* before write attempt, when no user RLOCK() or FLOCK() was performed and
* SET AUTOLOCK is currently active (the default) or the oRdd:CONCURRENCY-
* CONTROL instance is set. See description in the manual.
*
* AutoAPPEND() may be invoked manually instead of DBAPPEND() or APPE BLANK
* AutoRLOCK()  may be invoked manually instead of Rlock()
* AutoFLOCK()  may be invoked manually instead of Flock()
* AutoUNLOCK() may be invoked manually instead of UNLOCK
*
* DbAutoREOPEN() may be invoked manually to re-open current database and
*                indices in SHARE or EXCL mode using the same ALIAS and WA
* 
* These functions may be used for both procedural invocation (w/o argument)
* optionally prefaced with alias, and for Object oriented invocation (pass
* the RDD object as an argument).
*
* If METHODs named AutoRlock(), AutoFlock(), and AutoUnock() are available
* in the currently used RDD, these methods are preferred.
*
* These functions are available in object form in the FlagShip library.
* You may freely modify this file for your own purposes, e.g. modify the
* default BREAK invocation, activate the log protocol etc. But it is not
* recommended to change the general behaviour, since these functions ARE
* called from the default RDD and probably other drivers.
* It is recommended to modify a local copy of this file, not the original
* one stored in the <FlagShip_dir>/system directory.
*
* Compile      : FlagShip -na -c autolock.prg
* Link Linux   : FlagShip myappl*.prg autolock.o ....
* Link Windows : FlagShip myappl*.prg autolock.obj ....
*
*-----------------------------------------------------------------------------

#include "set.fh"
#include "rddsys.fh"

#define   PROTOCOL_ON                    /* disable, if not required      */
#define   WITH_BREAK                     /* enable, if required           */
#define   RDD_INFO_AVAILABLE             /* RDD specific                  */
#define   PREFER_RDD_METHOD              /* disable, if method in RDD     */

#define   MSRETRY_PERIOD 150             /* Milliseconds between retry    */
#define   PROT_TOOLONG   10              /* seconds for info msg to prot  */

#define   LOCK_RLOCK  1
#define   LOCK_FLOCK  2
#define   LOCK_UNLOCK 3
#define   LOCK_APPEND 4

#ifdef FS_WIN32                          /* Windows */
# define DEF_GETTIME()   Seconds()
# define DEF_WAITTIME(n) SleepMs(n)
#else
# ifdef FS_LINUX                         /* Linux */
#  define DEF_GETTIME()   Seconds()
#  define DEF_WAITTIME(n) SleepMs(n)
# else
#  define DEF_GETTIME()   SecondsCPU(4)   /* old Unix systems */
#  define DEF_WAITTIME(n) Sleep(max(int(n/1000),1))
# endif
#endif

#define OUT_MSG1 {"Try again", "Break to menu", "Ignore", "Quit"}
#define OUT_MSG2 "Trying autoRLOCK() for;"
#define OUT_MSG3 "Cannot perform autoRLOCK() for;"
#define OUT_MSG4 "Trying autoFLOCK() for;"
#define OUT_MSG5 "Cannot perform autoFLOCK() for;"
#define OUT_MSG6 "Trying autoUNLOCK for;"
#define OUT_MSG7 "Cannot perform autoUNLOCK for;"
#define OUT_MSG8 "Auto record lock"
#define OUT_MSG9 "Auto file lock"
#define OUT_MSG10 "Trying autoAPPEND for;"
#define OUT_MSG11 "Cannot perform autoAPPEND for;"
#define OUT_MSG12 "Auto append blank"
#define OUT_MSG13 "Trying re-open "
#define OUT_MSG14 "Could not re-open "
#define OUT_MSG15 {"Try again", "Abort"}

******************************************************************************
* AUTORLOCK([oRDD], [nWait]) -> retL
*
* "smart" RLOCK()
*
FUNCTION AUTORLOCK ([oRDD], [nWait]) AS LOGICAL
outProtocol("autoRlock() ")               // log protocol
return AutoWaitProc (oRDD, nWait, LOCK_RLOCK, OUT_MSG2, OUT_MSG3, OUT_MSG8)

******************************************************************************
* AUTOFLOCK([oRDD], [nWait]) -> retL
*
* "smart" FLOCK()
*
FUNCTION AUTOFLOCK ([oRDD], [nWait]) AS LOGICAL
outProtocol("autoFlock() ")               // log protocol
return AutoWaitProc (oRDD, nWait, LOCK_FLOCK, OUT_MSG4, OUT_MSG5, OUT_MSG9)

******************************************************************************
* AUTOUNLOCK([oRDD], [nWait]) -> retL
*
* "smart" COMMIT + UNLOCK
*
FUNCTION AUTOUNLOCK ([oRDD], [nWait]) AS LOGICAL
local nCommit := _GlobSetting(GSET_N_UNLOCK_COMMIT,0)
outProtocol("autoUnlock()")              // log protocol
 
if nCommit != 2 .and. valtype(oRDD) == "O"
   oRDD:COMMIT()                         // commit current database via RDD
elseif nCommit == 2
   DBCOMMITALL()                         // commit all databases
else
   DbCommit()                            // commit current database
endif
return AutoWaitProc (oRDD, nWait, LOCK_UNLOCK, OUT_MSG6, OUT_MSG7, "AutoUnlock")

******************************************************************************
* AUTOAPPEND([oRDD], [nWait]) -> retL
*
* "smart" DBAPPEND()
*
FUNCTION AUTOAPPEND ([oRDD], [nWait]) AS LOGICAL
outProtocol("autoAppend() ")              // log protocol
return AutoWaitProc (oRDD, nWait, LOCK_APPEND, OUT_MSG10, OUT_MSG11, OUT_MSG12)

******************************************************************************
* DBAUTOREOPen([lExcl], [nWait], [lReadOnly]) -> retL
*
* Automatically re-open current database and all assigned indices 
* in exclusive or shared mode in the same work area and same alias
*
* lExcl     = .T. exclusive, .F. shared mode (default) 
* nWait     = time-out retry seconds (default = 0: forever until success)
* lReadOnly = .T. read-only, .F. read-write (default)
*
FUNCTION DbAutoReopen (lExcl, [nWait], [lReadOnly]) AS LOGICAL
local oRdd, cDbfName, cRdd, aIdxName, cAlias, iSelect, iIdxSel
local cOutMsg, ok, ii, nStart, nTotalSec, iWaitMs

if valtype(lExcl) != "L"
   lExcl := .F.
endif
if valtype(nWait) != "N" .or. nWait < 1.0
   nWait := 0
endif
if valtype(lReadOnly) != "L"
   lReadOnly := .F.
endif 
cOutMsg := "DbAutoReopen() " + Dbf(,.T.) + " in " + if(lExcl,"EXCL","SHARED") + " mode"
 
if !used()
   outProtocol("# " + cOutMsg + " - CANOT REOPEN, no dbf open")  // log protocol
   return .F.                    // dbf is not open, exit
endif

outProtocol(cOutMsg)             // log protocol

oRdd     := DbObject()
cDbfName := Dbf(,.T.)            // dbf name with full path
aIdxName := IndexNames()         // all open indices for this dbf
cRdd     := oRdd:RddName         // current RDD
cAlias   := Alias()              // current alias
iSelect  := Select()             // current work area
iIdxSel  := IndexOrd()           // selected index no.
nStart   := DEF_GETTIME()        // start Seconds()
nTotalSec:= nStart 
iWaitMs  := _GlobSetting(GSET_N_AUTOLOCK_SLEEP, MSRETRY_PERIOD)
if iWaitMs < 10                  // wait for retry minim 10ms
   iWaitMs := MSRETRY_PERIOD     // or default of 150ms
endif
popupMessage(NIL, nStart)        // init popup

USE                              // close current dbf

/* re-open database
 */
SELECT (iSelect)
ok := DbUseArea(.F., cRdd, cDbfName, cAlias, !lExcl, lReadOnly) 
while !ok
   if !popupMessage(OUT_MSG13 + "database;" + cDbfName, 0) 
      DEF_WAITTIME(iWaitMs)      // wait 0.15s (or 1s)
   endif
   ok := DbUseArea(.F., cRdd, cDbfName, cAlias, !lExcl, lReadOnly) 
   if !ok .and. nWait > 0        // consider time-out?
      if nStart < DEF_GETTIME()  
         nStart := DEF_GETTIME()
      endif
      if DEF_GETTIME() - nStart > nWait   
         if userAction(OUT_MSG14 + "database;" + cDbfName, OUT_MSG15) == 1
            nStart := DEF_GETTIME()
         else 
            outProtocol("# " + cOutMsg + " - TIMEOUT, user choice = ABORT") 
            return .F.
         endif
      endif
   endif
enddo

/* re-open all assigned indices
 */
Neterr(.F.) 
if len(aIdxName) > 0                // indices available?
   OrdListClear()                   // yes, re-open all
   for ii := 1 to len(aIdxName)
      ok := DbSetIndex(aIdxName[ii], lExcl) 
      while !ok
         if !popupMessage(OUT_MSG13 + "index;" + aIdxName[ii], 0) 
            DEF_WAITTIME(iWaitMs)   // wait 0.15s (or 1s)
         endif
         ok := DbSetIndex(aIdxName[ii], lExcl) 
         if !ok .and. nWait > 0     // consider time-out?
            if nStart < DEF_GETTIME()
               nStart := DEF_GETTIME()
            endif
            if DEF_GETTIME() - nStart > nWait
               if userAction(OUT_MSG14 + "index;" + aIdxName[ii], OUT_MSG15) == 1
                  nStart := DEF_GETTIME()
               else 
                  outProtocol("# DbAutoReopen() " + Dbf(,.T.) + " index=" + aIdxName[ii] + " - TIMEOUT, user choice = ABORT") 
                  return .F.
               endif
            endif
         endif
      enddo
   next
   SET ORDER TO (iIdxSel)
   if Set(_SET_GOTOP) 
      DbGoTop()
   endif
endif

nTotalSec := DEF_GETTIME() - nTotalSec
if nTotalSec > PROT_TOOLONG
   outProtocol("! " + cOutMsg + " took " + ltrim(str(nTotalSec,6,1)) + " sec.")
endif
return .T.

******************************************************************************
* static (local working) functions follow
*
STATIC FUNCTION AutoWaitProc ([oRDD], [nWait], iType, cTxt1, cTxt2, cTxt3)
local nStartTime  := DEF_GETTIME()
local lockOk := .F., isMsg := .F., nUserIn := 0
local isObject := (valtype(oRDD) == "O")
local iWaitMs, nTotalSec, cFullName:= "", cFunctName := procname(1)

if nWait == NIL  
   nWait  := SET(_SET_AUTOLOCK)
endif

if nWait < -2                            // see SET AUTOLOCK
   return .F.                            // AutoLock not active
elseif nWait <= 0
   nWait := 99999999
endif
lockOk := if (isObject, oRDD:used, used())
if !lockOk
   return .F.
endif
iWaitMs  := _GlobSetting(GSET_N_AUTOLOCK_SLEEP, MSRETRY_PERIOD)
if iWaitMs < 10                          // wait for retry minim 10ms
   iWaitMs := MSRETRY_PERIOD             // or default of 150ms
endif
nTotalSec := DEF_GETTIME()

#ifdef PREFER_RDD_METHOD
   if !isobject                          // If the same named Method
      oRdd := dbobject()                 // is available in RDD,
   endif                                 // use it preferrably
   if isobjproperty(oRdd, 4, cFunctName) 
      return oRdd:&cFunctName()          // depends on the RDD functionality
   endif
#endif

popupMessage(NIL, nStartTime)            // init popup
while .T.                                // try again ?
   while .T.                             // loop for the lock retry
      do case
      case iType == LOCK_RLOCK
         if isObject
            lockOk := oRDD:RLOCK()       // object
         else
            lockOk := RLOCK()            // std. RLOCK()
         endif
      case iType == LOCK_FLOCK
         if isObject
            lockOk := oRDD:FLOCK()       // object
         else
            lockOk := FLOCK()            // std. FLOCK()
         endif
      case iType == LOCK_UNLOCK
         if isObject
            lockOk := oRDD:UNLOCK()      // object
         else
            UNLOCK                       // std. UNLOCK
            lockOk := .T.
         endif
      case iType == LOCK_APPEND
         if isObject
            lockOk := oRDD:Append()      // object
         else
            lockOk := DbAppend()         // std. Append Blank
         endif
      otherwise
         lockOk := .T.
      endcase
      if lockOk
         exit                            // successful
      endif

      if empty(cFullName)
#ifdef RDD_INFO_AVAILABLE
         if !isObject
            oRDD := DBOBJECT()           // get the RDD object
         endif
         cFullName := oRDD:INFO(DBI_FULLPATH)
         if empty(cFullName)
            cFullName := DBF()
         endif
#else
         cFullName := DBF()              // name of the database
#endif
      endif

      if DEF_GETTIME() < nStartTime      // over midnight
         nStartTime := DEF_GETTIME()     // new start time
         popupMessage(NIL, nStartTime)   // init popup
      endif
      if (DEF_GETTIME() - nStartTime) > nWait // SET AUTOLOCK time expired ?
         lockOk := .F.                   // yes
         exit
      endif

      if !popupMessage(cTxt1 + cFullName, nStartTime)  // display "trying.." message 
         DEF_WAITTIME(iWaitMs)           // wait 0.15s (or 1s)
      endif
   enddo                                 // end-of-loop for the lock retry

   if lockOk                             // successful ?
      exit
   endif

   nUserIn := userAction(cTxt2 + cFullName, OUT_MSG1)  // user choice
   if nUserIn == 1                       // try again
      nStartTime := DEF_GETTIME()
      outProtocol("# " + cTxt3 + " TIMEOUT, user choice = TRY AGAIN") 
      loop
   elseif nUserIn == 3                   // ignore
      outProtocol("# " + cTxt3 + " TIMEOUT, user choice = IGNORE") 
      exit
   elseif nUserIn == 4                   // quit
      outProtocol("# " + cTxt3 + " TIMEOUT, user choice = QUIT") 
      quit
   elseif nUserIn == 2                   // break
#ifdef WITH_BREAK
      if ISBEGSEQ()                      // Begin Sequence ... active?
         outProtocol("# " + cTxt3 + " TIMEOUT, user choice = BREAK") 
         BREAK cTxt3                     // yes, jump to the
      endif                              // next RECOVER or ENDSEQUENCE
#endif
   endif
   ?? chr(7)
enddo
nTotalSec := DEF_GETTIME() - nTotalSec
if nTotalSec > PROT_TOOLONG
   outProtocol("! " + cTxt3 + " took " + ;
         ltrim(str(nTotalSec,6,1)) + " sec." + ;
         if(lockOk,""," NOT") + " successful") 
elseif !lockOk  
   outProtocol("! " + cTxt3 + " NOT successful")
endif
return lockOk

******************************************************************************
* helper: print to log protocol, if desired
*
STATIC FUNCTION outProtocol (cTxt)
#ifdef PROTOCOL_ON
   local lOldCons, lOldPrint, cOldPrintTo, cPrintTo

   cPrintTo := _GlobSetting(GSET_C_AUTOLOCK_PROT, "")
   if empty(cPrintTo)                        // file name given?
      return                                 // no = no protocol
   endif
   
   lOldCons    := SET(_SET_CONSOLE)
   lOldPrint   := SET(_SET_PRINTER)
   cOldPrintTo := SET(_SET_PRINTFILE)
   
   SET CONSOLE OFF
   SET PRINTER TO (cPrintTo) ADDITIVE
   SET PRINTER ON
   ? date(),time(1), "EXE=" + execname(), "PID=" + ltrim(execpidnum()), ;
     cTxt, "src: " + procstack(2)
   SET PRINTER TO
   SET PRINTER OFF
   
   if !empty(cOldPrintTo)
      SET PRINTER TO (cOldPrintTo) ADDITIVE  // previous SET PRINTER TO...
   endif
   SET PRINTER (lOldPrint)                   // previous SET PRINTER on/off
   SET CONSOLE (lOldCons)                    // previous SET CONSOLE on/off
#endif
   return 

******************************************************************************
* helper: display pop-up info, if desired
*
STATIC FUNCTION popupMessage (cTxt, nStartTime)
local nStepMsg, nWaitMsg, nSeconds 
static nLastMsg := 0                // time of last msg

nStepMsg := _GlobSetting(GSET_N_AUTOLOCK_INFO,  5)  // display msg every...sec
nWaitMsg := _GlobSetting(GSET_N_AUTOLOCK_INFOWT,2)  // display msg for...sec

if nStepMsg <= 0 
   return .F.                       // no info pop-up desired
endif
if cTxt == NIL
   nLastMsg := nStartTime           // init popupMessage()
   return .T.
endif

nSeconds := DEF_GETTIME()           // current seconds()

if nSeconds < nLastMsg              // over midnight
   nLastMsg := nSeconds
endif
if nSeconds < (nLastMsg + nStepMsg) 
   return .F.                       // msg not displayed
endif
nLastMsg := nSeconds
InfoBox(cTxt, "Retry", nWaitMsg)    // display msg, auto time-out
return .T.

******************************************************************************
* helper: take user decission to continue, abort etc.
*
STATIC FUNCTION userAction (cTxt, aResponse)
return alert (cTxt, aResponse)

*** eof ***
