****************************************************************************
*
* smallRDD.prg == "smallRDD" RDD driver    (c) 1995..2005 by multisoft GmbH
*
* Release $Date: 2005/09/27 18:21:23 $
*
* Very small RDD driver for .dbf files, demonstrating how to use OOPs and
* the DBserver RDDs. Written in the FlagShip .prg language only, does NOT
* require any C knowledge. You may freely modify it for your own purposes,
* e.g. to access SQL databases by using e.g. Inline-C while calling ESQL
* of the SQL package.
*
* Heterogenous dbf access via object and/or command + function is supported
* here. There are two FlagShip functions (RDDannAlias and RDDannField) which
* connect (announce) the Object to the FlagShip functions and working area,
* see the oRdd:INIT() method below. If used, the complementary functions
* (RDDretAlias and RDDannField) retire this connenction when the database
* gets closed (see the oRdd:CLOSE() method below).
*
* This driver uses late .dbf update, which speeds up the process, since all
* fields of the same record are updated at once. The record is writen only
* if it was changed and a movement is issued, or if the last change is older
* than 5 seconds. You may disable this "late" updating by setting the #define
* WAIT_FOR_UPDATE to 0 (zero) or set the time difference to any higher value.
*
* Note, that the default DBFIDX driver uses other variables and more
* optimized techniques than the simplified ones here. For a full DBserver
* specification, see the *.c files in /usr/FSsrc/system/rdd4cb.tar.Z file.
*
* Methods: smallRDD:INIT()      --- internal, for object instantiation only
*          smallRDD:CLOSE()     <=> USE, DBCLOSEAREA()
*          smallRDD:SKIP ([n])  <=> SKIP [n], DBSKIP([n])
*          smallRDD:GOTO (n)    <=> GO TO n,  DBGOTO(n)
*          smallRDD:FIELDGET()  <=> FIELDGET(nPos|cName)
*          smallRDD:FIELDINFO() <=> DbStruct()
*          smallRDD:FIELDPUT()  <=> FIELDUPT(nPos|cName, value)
*          smallRDD:DBSTRUCT()  <=> DBSTRUCT()
*          smallRDD:NOIVARGET() --- run-time access to oRdd:<field>
*          smallRDD:NOIVARPUT() --- run-time assign to oRdd:<field>
* Access:  smallRDD:USED        <=> USED()
*          smallRDD:ALIAS       <=> ALIAS()
*          smallRDD:DELETED     <=> DELETED()
*          smallRDD:RECNO       <=> RECNO()
*          smallRDD:RECCOUNT    <=> RECCOUNT(), LASTREC()
*          smallRDD:RECSIZE     <=> RECSIZE()
*          smallRDD:EOF         <=> EOF()
*          smallRDD:BOF         <=> BOF()
*          smallRDD:FCOUNT      <=> FCOUNT()
*          smallRDD:INDEXCOUNT  <=> INDEXCOUNT() 
*          smallRDD:SHARED      <=> SHARED()
* Assign:  smallRDD:ALIAS       --- invoked from USE and DBUSEAREA()
* Export:  smallRDD:lEOF        --- (not required at all)
*          smallRDD:lBOF        --- (not required at all)
*
* Compile: FlagShip -na -c -w1 -w3 -w4 smallRDD.prg
* Include: rename the 'reposit.fh' produced by FScomp to 'smallRDD.fh' and
*            #include "smallRDD.fh" to your applic (myappli*.prg)
*          for prototyping, if used in your application (higly recommended).
* Link   : FlagShip myappli*.prg smallRDD.o
*
* Usage a) USE mydbf NEW VIA smallRDD
*          if USED()
*            GOTO x, SKIP, ? <field>, lEOF(), lBOF(), ...
*            REPLACE <field> ..., FIELDPUT(field,var), ...
*            USE
*          endif
*
* Usage b) #include "smallRDD.fh"          // recommended but not necessary
*          Local oDbf AS smallRDD          // recommended but not necessary
*          oDbf := SMALLRDD {"mydbf",,,,,.T.}
*          if oDbf:USED
*             oDbf:GOTO(x), oDbf:SKIP(), QOUT(oDbf:field, oDbf:lEOF, oDbf:lBOF)
*             oDbf:field := ..., oDbf:FIELDPUT(field,val), ...
*             oDbf:CLOSE()
*          endif
*
* Usage c) all combinations of (a) and (b) are supported, e.g.:
*          #include "smallRDD.fh"          // recommended
*          Local oDbf AS smallRDD          // recommended
*          USE mydbf NEW VIA smallRDD
*          if USED()
*             oDbf := DBOBJECT()
*             oDbf:GOTO(x), SKIP, ? oDbf:field, oDbf:lEOF, BOF()
*             DBCLOSEAREA()
*          endif
*
* Example: srddtest.prg
*
****************************************************************************

#include "smallrdd.fh"              /* class prototypes */

#include "fileio.fh"
#include "error.fh"
// #include "FSerrmsg.h"            /* only if _rt_error(..) is used */

#define WAIT_FOR_UPDATE 5           /* set to 0 to write immediately   */
#define CONSIDER_SET_DELETED        /* SET DELETED is considered       */
#define TEST_ONLY                   /* activate for test purposes      */

#define ERR_NoDbfUsed   1
#define ERR_ReplOnEof   2
#define ERR_ReadError   3

#define OFFSET_DBF_VERS 0           /* offset of .dbf database type    */
#define OFFSET_UPD_DATE 1           /* LUPDATE() offset in .dbf header */
#define OFFSET_RECCOUNT 4           /* offset of record count          */
#define OFFSET_HDRLEN   8           /* offset of header size           */
#define OFFSET_RECLEN   10          /* offset of record size           */
#define OFFSET_1STFIELD 32          /* offset of 1st field descriptor  */
#define SIZE_HEADFIELD  32          /* size of field descr in header   */


****************** announce RDD **********************************************
INIT procedure _smallrdd
   REQUEST smallrddNEW        // will be created automatically
   AnnNewRdd ("SmallRDD")     // in rddsys.prg, announce for RddList()
RETURN

****************** class creator *********************************************
CLASS smallRDD INHERIT DataServer 

****************** class properies *******************************************
*
******************************************************************************
* oRdd:INIT(cFileName, [lShare], [lReadOnly],
*           [p4], [p5], [lNew]) -> self               internally used Method
*
* cFilename : name of the .dbf file, optionally with path and extention
* lShare    : ignored here, allways assumed .F. (exclusive)  
* lReadOnly : ignored here, allways assumed .F. (read/write) 
* p4        : rdd name, ignored here
* p5        : alias name, ignored here
* lNew      : if set .F. or is NIL, the current working area is used. 
*				  If specified .T., a new working area is used, equivalent to 
*             the NEW clause of USE.
*
* Called internally on object instantiation or from USE, DBUSEAREA().
* Opens the database (always .dbf), fs_set() and SET PATH is considered.
* Fills object instances. Notifies FlagShip about the alias/work area and
* fields in use for heterogenous program usage.
*
METHOD init(cName, lShare, lReadonly, dummy1, dummy2, lNew) CLASS smallRDD
   Local ii, jj AS int
   if valtype(lNew) == "L" .and. lNew
      SELECT 0
   else
      if self:lUsed .and. self:iHandle > 0   // open by this RDD?
         self:close()
      elseif Used()                          // open by other RDD?
         USE
      endif
   endif
   self:iHandle := 0
   NETERR (.T.)                                   // set NetErr()
   if valtype(cName) != "C" .or. empty(cName)     // param checking
      self:lUsed := .F.
      return self
   endif
   self:lUsed := self:rddOpenDBF (@cName)             // open database file
   if self:lUsed
      self:lUsed := self:fillRddInstances (cName)     // see method below
   endif
   if self:lUsed
      self:cAlias := RDDannAlias (self,    ;          // returns actual
                              self:cFullFileName)     // alias name
      if empty (self:cAlias)
         alert ("smallRDD:INIT() cannot create/modify alias '" + ;
                self:cAlias + "'")
         self:close()
         return self
      endif
      self:cName := cName
      jj := len(self:aFields)
      self:aFldPtr := ARRAY(jj)                    // create carr.array
      for ii := 1 to jj                            // Announce _FIELD Vars:
         RDDannField (self:aFields[ii,DBS_NAME], ; // -field name
                      ii,                        ; // -field pos
                      self:aFields[ii,DBS_TYPE], ; // -field type
                      1,                         ; // -late field fill
                      self:aFields[ii,DBS_LEN ], ; // -field length
                      self:aFields[ii,DBS_DEC ], ; // -field deci length
                      self:aFldPtr)                // -carrying array
      next
      self:cOneRecBuff := space(self:iRecLen)      // assign buffer
      self:goto(1)                                 // read first record
      NETERR (.F.)                                 // reset NetErr()
   endif
   self:nChangeTimeme:= 0                          // init instance
   self:lChanged   := .F.                          // init instance
#ifdef TEST_ONLY
   self:displayInstances(.T.)
#endif
   return self

******************************************************************************
* oRdd:CLOSE() -> lSuccess                                              Method
*
* Close the database, free the assigned work area, alias and fields
* Equivalent to USE or DBCLOSEAREA()
*
METHOD close() CLASS smallRDD
   Local ok := .F., ii
   if self:lUsed .and. self:iHandle > 0             // is the .dbf open ?
      if self:lChanged                              // update record if changed
         self:WriteRecord(.T.)
      endif
      FSEEK  (self:iHandle, 0, FS_START)            // try to unlock
      FLOCKF (self:iHandle, FF_UNLOCK)              // the whole .dbf
      FSEEK  (self:iHandle, 0, FS_END)              // required for some systems
      ok := FCLOSE (self:iHandle)                   // close
   endif
   for ii := 1 to len(self:aFldPtr)                 // free _FIELD vars
      RDDretField (self, self:aFields[ii,DBS_NAME]) // in FlagShip
   next
   self:cAlias       := ""                          // reset instances
   self:aFields      := {}
   self:lUsed        := .F.
   self:iHandle      := 0
   RDDretAlias (self)                               // free Alias in FS
   return ok

******************************************************************************
* oRdd:USED -> lUsed                                                    Access
* equivalent to USED()
*
ACCESS used CLASS smallRDD
   return self:lUsed

******************************************************************************
* oRdd:lEOF  -> lEndOfFile                                              Access
* equivalent to EOF()
*
ACCESS eof CLASS smallRDD
   return self:lEof

******************************************************************************
* oRdd:lBOF  -> lBeginOfFile                                            Access
* equivalent to BOF()
*
ACCESS bof CLASS smallRDD
   return self:lBof

******************************************************************************
* oRdd:Fcount -> nFields                                                Access
* equivalent to FCOUNT()
*
access fcount CLASS smallRDD		
   return len(self:aFields)

******************************************************************************
* oRdd:LASTREC -> nRecords                                              Access
* oRdd:RECCOUNT -> nRecords                                             Access
* equivalent to RECCOUNT(), LASTREC()
*
access lastrec CLASS smallRDD		
return self:RecCount

ACCESS reccount CLASS smallRDD
   return INT2NUM(self:iLastRec)        // for ValType() == 'N'

******************************************************************************
* oRdd:RECSIZE  -> nRecordSize                                          Access
* equivalent to RECSIZE()
*
ACCESS recsize CLASS smallRDD
   return INT2NUM(self:iRecLen)        // for ValType() == 'N'

******************************************************************************
* oRdd:RECNO <-> nRecno                                          Access/Assign
* equivalent to RECNO() or DBGOTO(nRecno)
*
ACCESS recno CLASS smallRDD
   return INT2NUM(self:iRecNo)        // for ValType() == 'N'

ASSIGN recno(value) CLASS smallRDD
   if valtype(value) $ "NI"
      self:GOTO(value)
   endif
   return INT2NUM(self:iRecNo)        // for ValType() == 'N'

******************************************************************************
* oRdd:Shared -> no (simulated only, no locking here)                               
*
access shared CLASS smallRDD		
return .F.

******************************************************************************
* oRdd:IndexCount -> 0 in this RDD                              
*
access indexcount CLASS smallRDD	
return 0

******************************************************************************
* oRdd:ALIAS <-> cAlias                                          Access/Assign
*
* equivalent to ALIAS().
* Caution: do NOT change the alias when using hybrid programming!
*
ACCESS alias CLASS smallRDD
   return self:cAlias

ASSIGN alias(value) CLASS smallRDD
   LOCAL cAlias

   if valtype(value) == "C"
      cAlias := RDDannAlias (self, value)        // returns actual alias name
      if empty (cAlias)
         alert ("smallRDD:ALIAS() cannot assign alias '" + cAlias + "'")
         return self:cAlias
      endif
     self:cAlias := cAlias
   endif
   return self:cAlias

******************************************************************************
* oRdd:DELETED -> lDeleted                                              Access
* equivalent to DELETED()
*
ACCESS deleted CLASS smallRDD
   return left(@self:cOneRecBuff,1) $ "*TtYy"

******************************************************************************
* oRdd:FIELDGET(nFieldPos|cFieldName) -> value                          Method
*
* Equivalent to FIELDGET() function.
* This method is also automatically invoked on <value> := _FIELD-><name>
* according to the RDDannField() specification in oRdd:init()
*
METHOD FieldGet(ncField) CLASS smallRDD
   Local nFieldNo := 0, cFieldName, value, cType
   self:RereadRecord()
   if EMPTY(self:lUsed)
      return NIL
   endif
   if valtype(ncField) == "C"
      cFieldName := left(upper(alltrim(ncField)), 10)
      nFieldNo := self:preGetCheck (cFieldName, "Field", "accessed by FIELDGET()")
   elseif valtype(ncField) $ "NI"
      nFieldNo := ncField
   else
      return NIL                                      // wrong param
   endif
   if nFieldNo <= 0 .or. nFieldNo > len(self:aFldPtr)
      return NIL
   endif
   ctype := self:aFields[nFieldNo,DBS_TYPE]
   value := self:getFieldValue (nFieldNo, cType)      // field -> variable
   if value == NIL
      ALERT ("Unknown/unsupported type '" + ctype + "' of Field" + ;
             if (valtype(ncField) == "C", "-Name '" + cFieldName + "'", ;
                 "-Number " + ltrim(str(nFieldNo)) ) + ;
             ";of the class smallRDD accessed in;"   + ;
             procname(1) + "/" + alltrim(str(procline(1))))
      return NIL
   endif
   AssignFldValue (@self:aFldPtr[nFieldNo], value)    // don't change SubType
   return @self:aFldPtr[nFieldNo]                     // varSubType = field

******************************************************************************
* oRdd:FIELDPUT(nFieldPos|cFieldName, value) -> lSuccess                Method
*
* Equivalent to FIELDPUT() function.
* This method is also automatically invoked on _FIELD-><name> := <value>
* according to the RDDannField() specification in oRdd:init()
*
METHOD FieldPut(ncField, value) CLASS smallRDD
   Local nFieldNo := 0, cFieldName, cType, ok
   if EMPTY(self:lUsed) .or. self:lEof
      self:displErr (if (EMPTY(self:lUsed), ERR_NoDbfUsed, ERR_ReplOnEof))
      return .F.
   endif
   if valtype(ncField) == "C"
      cFieldName := left(upper(alltrim(ncField)), 10)
      nFieldNo := self:preGetCheck (cFieldName, "Field", ;
                                    "accessed by FIELDPUT()")
   elseif valtype(ncField) $ "NI"
      nFieldNo := ncField
   else
      return .F.                                      // wrong param
   endif
   if nFieldNo <= 0 .or. nFieldNo > len(self:aFldPtr)
      return .F.
   endif
   ctype := self:aFields[nFieldNo,DBS_TYPE]
   ok := self:putFieldValue (nFieldNo, value)         // variable -> fieldbuff
   if !ok
      ALERT ("Unknown/unsupported/different type '" + ctype + "' of Field" + ;
             if (valtype(ncField) == "C", "-Name '" + cFieldName + "'", ;
                 "-Number " + ltrim(str(nFieldNo)) ) + ;
             ";replaced by variable type '" + valtype(value) + "'" + ;
             ";of the class smallRDD accessed in;"   + ;
             procname(1) + "/" + alltrim(str(procline(1))))
      return .F.
   endif
   if cType == "C"
      AssignFldValue (@self:aFldPtr[nFieldNo],        // assign to _FIELD->..
                      PADR (value, self:aFields[nFieldNo,DBS_LEN]))
   else                                               // but don't change
      AssignFldValue (@self:aFldPtr[nFieldNo], value) // the var's SubType
   endif
   self:lChanged := .T.                               // record is modified
   self:WriteRecord(.F.)                              // write on request
   self:RereadRecord()
   return .T.

******************************************************************************
* oRdd:FIELDINFO(nInfo, nFieldPos|cFieldName) -> value                  Method
*
METHOD FieldInfo(nWhat, ncField) CLASS smallRDD
	local cVar		 as char
	local iPos := 0 as intvar
	if valtype(nWhat) != "N" .or. nWhat < 1 .or. nWhat > 4
		return NIL
	endif
	if valtype(ncField) == "C"
   	cVar := upper(trim(ncField))
   	iPos := ASCAN (self:aFields, {|x| cVar == upper(trim(x[DBS_NAME])) })
	elseif valtype(ncField) == "N"
		iPos := ncField
	endif
   if iPos > 0 .and. iPos <= len(self:aFields)
		return self:aFields[iPos, nWhat]
	endif
   return NIL
	
******************************************************************************
* oRdd:FIELDNAME(nFieldPos) -> cName                                    Method
*
* Equivalent to FIELNAME() function.
*
METHOD FieldName(nField) CLASS smallRDD
   if valtype(nField) != "N" .or. nField < 1 .or. nField > len(self:aFields)
      alert("Invalid field number " + ltrim(nField) + " (1.."+ltrim(len(self:aFields))+")")
      return ""
   endif
   return upper(self:aFields[nField,1])
   
******************************************************************************
* oRdd:NoIvarGet(cVarName) -> value                       internal used Method
*
* called from the FS runtime system on unknown oRdd:<name> access, e.g.
* by using assignment value := oRdd:myFieldName 
*
METHOD NoIvarGet(cVar) CLASS smallRDD
   Local nField
   nField := self:preGetCheck (cVar, "Field/Variable", "accessed")
   if nField == 0
      return .F.
   endif
   return @ self:FieldGet(nField)

******************************************************************************
* oRdd:NoIvarPut(cVarName, value) -> lSuccess             internal used Method
*
* called from the FS runtime system on unknown oRdd:<name> assign, e.g.
* by using assignment oRdd:myFieldName := value
*
METHOD NoIvarPut(cVar, value) CLASS smallRDD
   Local nField
   nField := self:preGetCheck (cVar, "Field/Variable", "assigned")
   if nField == 0
      return .F.
   endif
   return self:FieldPut(nField, @value)

******************************************************************************
* oRdd:SKIP([nNoOfRecords]) -> lSuccess                                 Method
*
* Equivalent to SKIP [n] command and DBSKIP([n]) function
* Note: SET DELETED is considered, but not SET FILTER
*
METHOD skip(nCount) CLASS smallRDD
   Local iSign, new           AS IntVar
   Local iSoll, iSkipped := 0 AS INT

   if valtype(nCount) != "N"
      nCount := 1
   endif
   if !self:preSkipCheck()                            // param, used, empty ?
      RETURN .F.
   endif
   if self:lChanged                                   // curr.record changed?
      self:WriteRecord(.T.)                           // yes, write first
   endif
   self:lBof := self:lEof := .F.

#ifdef CONSIDER_SET_DELETED
   if !SET(_SET_DELETED)                              // SET DELETED OFF ?
#endif
      new := FSEEK (self:iHandle, (nCount * self:iRecLen), FS_RELATIVE)
      self:iRecNo := self:iRecNo + nCount
      self:ReadRecord()                               // read record in buff
      if ! self:postSkipCheck (new)                   // set EOF or BOF ?
         return .F.
      endif
#ifdef CONSIDER_SET_DELETED
   else                                               // SET DELETED ON
      iSign := NUM2INT(ncount / abs(ncount))          // = +/- 1
      iSoll := NUM2INT(abs(nCount))
      while !self:lEof .and. !self:lBof .and. iSkipped < iSoll
         new := FSEEK (self:iHandle, (iSign * self:iRecLen),
                       FS_RELATIVE)                   // skip +/- 1 record
         self:iRecNo := self:iRecNo + iSign
         self:ReadRecord()                            // read record in buff
         self:postSkipCheck (new)                     // set EOF or BOF
         if ! (self:Deleted .or. SET(_SET_DELETED))
            iSkipped++                                // count not deleted
         endif                                        // records only
      enddo
      if self:lBof .and. self:Deleted .and. SET(_SET_DELETED)
         while self:Deleted .and. !self:lEof
            new := FSEEK (self:iHandle, self:iRecLen, FS_RELATIVE)
            self:iRecNo++
            self:ReadRecord()                         // read record in buff
            self:postSkipCheck (new)
         enddo
      endif
      return (iSkipped == iSoll)
   endif
#endif
   return .T.

******************************************************************************
* oRdd:GOTO (nRecNo) -> lSuccess                                        Method
*
* equivalent to GO TO n command and DBGOTO(n) function
*
METHOD goto(nRec) CLASS smallRDD
   Local new AS intVar
   if !(valtype(nRec) $ "NI") .or. ;
      nRec < 0 .or. nRec > (self:iLastRec +1) .or. ;
      ! self:preSkipCheck()                           // param, used, empty ?
      RETURN .F.
   endif
   if self:lChanged                                   // curr.record changed?
      self:WriteRecord(.T.)                           // yes, write first
   endif
   new := FSEEK (self:iHandle, self:iStartPos + ((nRec -1) * self:iRecLen), FS_START)
   self:lBof := self:lEof := .F.                      // reset EOF and BOF
   self:iRecNo := nRec
   self:ReadRecord()                                  // read record in buff
   if ! self:postSkipCheck (new)                      // set EOF or BOF ?
      return .F.
   endif
   return .T.

******************************************************************************
* oRdd:DBSTRUCT() -> array                                              Method
*
* equivalent to DBSTRUCT() function
*
METHOD DBstruct() CLASS smallRDD
   return self:aFields

*-----------------------------------------------------------------------------
* display error message. You may use _rt_error(num,string) instead of
* Alert() to produce real RTE message. If so, enable #include "FSerrmsg.h"
*
PROTECT METHOD displErr (nMsg, [var1], [var2])  CLASS smallRDD
   Local cWA := ltrim(str(SELECT()))
   do case
   case nMsg == ERR_NoDbfUsed
//    _rt_error(RTNODBF, "No database open in WA " + cWA)
      ALERT ("No database open in WA " + cWA )
   case nMsg == ERR_ReplOnEof
//    _rt_error(RTILLEOF, "Cannot replace on EOF() in WA " + cWA)
      ALERT ("Cannot replace on EOF() in WA " + cWA)
   case nMsg == ERR_ReadError
//    _rt_error(IOREAD, "Read error " + in WA " + cWA)
      ALERT ("Read error " + ltrim(str(FERROR())) + ;
             ": in WA " + cWA + ". Only " + ltrim(str(var1)) + ;
             " instead of " + ltrim(str(var2)) + " bytes read")
   endcase
   return NIL

*-----------------------------------------------------------------------------
* open the .dbf file, display error if not found, store handle number if ok
*
PROTECT METHOD rddOpenDBF (cName)    CLASS smallRDD                    
   Local ok := .F., cFullName := cName, iErr
   iHandle := 0
   if !(upper(right(cName, 4)) == ".DBF")
      cFullName += ".dbf"                        // add default extention
   endi
   if !file(cFullName)
      cFullName := TruePath (cFullName)          // incl.FS_SET() and SET PATH
   endif
   if file(cFullName)
      iHandle := FOPEN (cFullName, FO_READWRITE) // return handle number
   endif
   if iHandle <= 0                               // success ?
      iErr := FERROR()
      iErr := if (iErr == 0, EX_ENOENT, iErr)
      ALERT ("Database '" + cFullName + "' not found,;ErrorCode = " + ;
             ltrim(str(iErr)))
      return .F.
   endif
   self:iHandle := iHandle                       // used by FREAD() etc.
   cName := cFullName                            // return full file name
   return .T.

*-----------------------------------------------------------------------------
* check before skip if WA is open, reset flags
*
PROTECT METHOD preSkipCheck() CLASS smallRDD                       
   if EMPTY(self:lUsed)
      self:displErr (ERR_NoDbfUsed)
      return .F.
   endif
   if self:iLastRec == 0
      self:lBof := self:lEof := .T.
      return .F.
   endif
   return .T.

*-----------------------------------------------------------------------------
* check if <new> record is before first or behind last record, fix if required
*
PROTECT METHOD postSkipCheck(new AS intVar) CLASS smallRDD         
   if new < self:iStartPos .or. self:iRecNo < 0
      FSEEK (self:iHandle, self:iStartPos, FS_START)
      self:lBof     := .T.
      self:iRecNo := 1
      self:ReadRecord()                               // read record in buff
      return .F.
   endif
   if new >= self:iEndPos .or. self:iRecNo > self:iLastRec
      self:lEof     := .T.
      self:iRecNo := self:iLastRec +1
      self:ReadRecord()                               // read record in buff
      return .F.
   endif
   return .T.

*-----------------------------------------------------------------------------
* extraxt field value from internal record buffer
*
PROTECT METHOD getFieldValue (nFieldNo, type)  CLASS smallRDD      
   DO CASE
   CASE type == "C"
      return substr(self:cOneRecBuff, self:aRelPos[nFieldNo],self:aFields[nFieldNo,DBS_LEN])
   CASE type == "L"
      return (substr(self:cOneRecBuff, self:aRelPos[nFieldNo], 1) $ "TtYy")
   CASE type == "N"
      return val(substr(self:cOneRecBuff, self:aRelPos[nFieldNo], ;
                 self:aFields[nFieldNo,DBS_LEN])) //, self:aFields[nFieldNo,DBS_DEC])
   CASE type == "D"
      return stod(substr(self:cOneRecBuff, self:aRelPos[nFieldNo], 8))
   OTHERWISE
      return NIL
   ENDCASE

*-----------------------------------------------------------------------------
* convert <value> to string, normalize and store it in internal record buffer
*
PROTECT METHOD putFieldValue (nFieldNo, value)  CLASS smallRDD     
   Local cType := valtype(value), cBuff
   if nFieldNo <= 0 .or. nFieldNo > len(self:aFields) .or. ;
      !(cType == self:aFields[nFieldNo,DBS_TYPE])
      return .F.                                      // wrong type
   endif
   DO CASE
   CASE ctype == "C"
      cBuff := PADR (value, self:aFields[nFieldNo,DBS_LEN])
   CASE ctype == "L"
      cBuff := if (value, "T", "F")
   CASE ctype == "N"
      cBuff := str(value, self:aFields[nFieldNo,DBS_LEN], ;
                   self:aFields[nFieldNo,DBS_DEC])
   CASE ctype == "D"
      cBuff := dtos (value)
   ENDCASE
   if cbuff != NIL                                    // supported type?
      self:cOneRecBuff := stuff(self:cOneRecBuff, self:aRelPos[nFieldNo], ;
             self:aFields[nFieldNo,DBS_LEN], cBuff)   // replace record buff
      self:lChanged := .T.
      return .T.
   endif
   return .F.

*-----------------------------------------------------------------------------
* check if <cVar> is a valid field name
*
PROTECT METHOD preGetCheck(cVar, cText1, cText2) CLASS smallRDD    
   Local VarPos
   cVar := upper(trim(cVar))
   VarPos := ASCAN (self:aFields, {|x| cVar == upper(trim(x[DBS_NAME])) })
   if VarPos == 0
      if !empty(cText1)
         ALERT ("Unknown " + cText1 + " '" + cVar + ;
                "';of the class smallRDD " + cText2 + " in;" + ;
                procname(2) + "/" + alltrim(str(procline(2))))
      endif
   endif
   return VarPos

*-----------------------------------------------------------------------------
* read one record from .dbf file to internal RDD record buffer
* the file offset remains at the current position
*
PROTECT METHOD ReadRecord() CLASS smallRDD                         
   Local iCurrPos, iRead

   if self:lEof
      self:cOneRecBuff := space(self:iRecLen)
      return NIL
   endif
   iCurrPos := FSEEK (self:iHandle, 0, FS_RELATIVE)        // current file posit
   if iCurrPos >= self:iEndPos
      self:lEof := .T.
      self:cOneRecBuff := space(self:iRecLen)
      return NIL
   endif
   iRead := FREAD(self:iHandle, @self:cOneRecBuff, self:iRecLen)  // read record
   FSEEK (self:iHandle, iCurrPos, FS_START)                       // go back
   if iRead != self:iRecLen
      self:displErr(ERR_ReadError, iRead, self:iRecLen)
      self:cOneRecBuff := space(self:iRecLen)
   endif
   return NIL

*-----------------------------------------------------------------------------
* ensure you are using current data. Relevant only on SHARED database
*
PROTECT METHOD RereadRecord() CLASS smallRDD                      
// return self:ReadRecord()
   return NIL     // not required here, since exclusive
   
*-----------------------------------------------------------------------------
* write the current record to .dbf file
*
PROTECT METHOD WriteRecord(lMust) CLASS smallRDD                   
   Local iCurrPos, currSeconds := seconds()
   Local cUpdDate

   if !lMust                                           // write required ?
      if !self:lChanged .or. WAIT_FOR_UPDATE <= 0      // no, changes pending?
         return NIL                                    // no, do nothing
      endif
//      if self:nChangeTimeme > 0 .and. ;
//         self:nChangeTimeme <= currSeconds .and. ;
//         (currSeconds-self:nChangeTimeme) < WAIT_FOR_UPDATE  // don't write, if
//         return NIL                                    // the last update
//      endif                                            // was within 5 sec
   endif
   iCurrPos := FSEEK (self:iHandle, 0, FS_RELATIVE)      // current file posit
   FWRITE (self:iHandle, self:cOneRecBuff, self:iRecLen) // write record
   cUpdDate := chr(year(date() - 1900)) + ;
               chr(month(date())) + chr(day(date()))
   FSEEK  (self:iHandle, OFFSET_UPD_DATE, FS_START)      // update the .dbf
   FWRITE (self:iHandle, cUpdDate, 3)                    // modification date
   FSEEK  (self:iHandle, iCurrPos, FS_START)             // go back
   self:lChanged    := .F.                               // no changes pending
   self:nChangeTimeme := currSeconds                     // time of last update
   return NIL

*-----------------------------------------------------------------------------
* init the internal RDD data
*
PROTECT METHOD fillRddInstances(cName) CLASS smallRDD              
   Local iPos, iFld, iZero, iFieldCount := 0 AS IntVar
   Local ii, iBuff, iRecPos                  AS LONG
   Local cField as char
   Local cBuff := space(OFFSET_1STFIELD)

   self:lEof := self:lBof := .T.                   // set EOF() and BOF()
   self:cFullFileName := cName                     // set FileName instance
   cName := strtran(cName, "\", "/")               // normalize file name

   FSEEK (self:iHandle, 0, FS_START)               // read the first 16
   FREAD (self:iHandle, @cBuff, 16)                // bytes of .dbf header
   ii := strpeek(@cbuff, OFFSET_DBF_VERS +1)
   if ii != 3                                      // 0x03 = regular .dbf
      if ii == 131                                 // 0x83 = .dbf and .dbt
         ALERT ("Warning: .dbt and memo fields not supported here")
      else
         ALERT ("Error: '" + self:cFullFileName + ;
                "' is not a regular .dbf database")
         FCLOSE (self:iHandle)                     // close database
         self:lUsed := .F.
         return .F.
      endif
   endif
   self:iRecLen := BIN2I(substr(@cbuff, OFFSET_RECLEN +1, 2))  // incl.deleted,
   if self:iRecLen < 0                                         // in bytes
      self:iRecLen = 65535 + self:iRecLen                      // correct, if
   endif                                                       // signed
   self:iLastRec  := BIN2L(substr(@cbuff, OFFSET_RECCOUNT +1, 4)) // 0..n records
   self:iStartPos := BIN2I(substr(@cbuff, OFFSET_HDRLEN   +1, 2)) // 1st valid
   self:iEndPos   := self:iStartPos + (self:iLastRec * self:iRecLen) // last valid

   FSEEK (self:iHandle, 0, FS_START)                           // try flock
   ii := FLOCKF (self:iHandle, FF_READWRITE, 2147483648)       // 2 GBytes
   if ii != 0                                                  // or at least
      ii := FLOCKF (self:iHandle, FF_READWRITE)                // the database
   endif
   if ii != 0
      if ALERT ("Exclusive lock on '" + cName + "' failed.;Continue ?", ;
                {"No", "Yes"}) != 2
         FCLOSE (self:iHandle)                                 // close dbf
         self:lUsed := .F.
         return .F.
      endif
   endif

   iFld := NUM2INT((self:iStartPos -1 - OFFSET_1STFIELD) / ;
                    SIZE_HEADFIELD)                           // no of fields
   self:aFields := array(iFld)                                // dbstruct()
   self:aRelPos := array(iFld)                                // rel.posit.fld
   iBuff := (iFld * SIZE_HEADFIELD) +1                        // incl eohead
   cBuff := space(iBuff)
   FSEEK (self:iHandle, OFFSET_1STFIELD, FS_START)            // read field
   FREAD (self:iHandle, @cBuff, iBuff)                        // descriptors
*
*  FlagShip uses the last byte in header for consistency check
*
*  if strpeek(@cBuff, iBuff) # 13
*     ALERT ("Warning: possible .dbf header corruption")      // 0x0d = eohead
*  endif

   iPos    := 1                           // relat.position in header
   iRecPos := 2                           // since coOneRecBuf[1]=deleted flag
   for ii := 1 to iFld
      cField := substr(@cBuff, ipos, 10)
      if at(chr(0),cField) > 0            // field name is \0 terminated
         cField := left(cField, at(chr(0),cField) -1)
      endif
      self:aFields[ii] := {upper(cField), ;                       // fld name
                       upper(substr(@cBuff, ipos+11, 1)), ;       // fld type
                       strpeek(@cBuff, ipos+16),          ;       // fld len
                       strpeek(@cBuff, ipos+17) }                 // fld deci
      if self:aFields[ii,2] == "C" .and. self:aFields[ii,4] > 0   // correct the
         self:aFields[ii,3] := self:aFields[ii,4] * 256   ;       // char size
                         + self:aFields[ii,3]                     // 1..65534
         self:aFields[ii,4] := 0
      endif
      if (iZero := at(chr(0),self:aFields[ii,1]) ) > 0            // name in database 
         self:aFields[ii,1] := left(self:aFields[ii,1], iZero -1) // is zero terminated
      endif
      self:aRelPos[ii] := iRecPos                                 // rel.fld posit
      iRecPos += self:aFields[ii,3]                               // and increment
      iPos += SIZE_HEADFIELD
   next
   return .T.

*-----------------------------------------------------------------------------
* for test/debugging purposes only
*
METHOD displayInstances(lPrint) CLASS smallRDD
#ifdef TEST_ONLY
   Local lPrintOnly := if (valtype(lPrint) == "L", lPrint, .F.)
   if lPrintOnly
      Local isprint := SET (_SET_PRINTER)
      Local iscons  := SET (_SET_CONSOLE)
      SET PRINTER ON
      SET CONSOLE OFF
   endif
// #ifdef FlagShip5
//    if lPrintOnly
//       _DisplObj(self, "self")
//    else
//       _DisplObjErr(self, "self")   // on stderr
//    endif
// #else
   ? "---- smallRDD instances at " + procname(1) +"/"+ ltrim(str(procline(1)))
   ? "lUsed,    oRdd:Used   ", self:lUsed, self:Used
   ? "[cAlias], oRdd:Alias  ", "["+self:cAlias+"]", self:Alias
   ? "eof, bof              ", self:lEof, self:lBof
   if self:lUsed
      ? "[self:cFullFileName]       ", "[" + self:cFullFileName + "]"
      if len(self:cOneRecBuff) <= 55
         ? "[cOneRecBuf]          ", "[" + self:cOneRecBuff + "]"
      else
         ? "[cOneRecBuf]          ", "[" + left(self:cOneRecBuff,44) + ;
                                      "]...[" + right(self:cOneRecBuff,5) + "]"
      endif
      ? "iRecLen, oRdd:RECSIZE ", self:iRecLen,  self:Recsize
      ? "iRecNo,  oRdd:RECNO   ", self:iRecNo,   self:Recno
      ? "iLastRec,oRdd:RECCOUNT", self:iLastRec, self:RECCOUNT
      ? "iHandle               ", self:iHandle
      ? "iStartPos, ioEndPos   ", self:iStartPos, self:iEndPos
      ? "nChangeTime, loChanged", self:nChangeTimeme, self:lChanged

      ? "Dbstruct:"
      aeval (self:aFields, {|x| qout(padr("["+x[1]+"]",12), x[2], x[3], x[4])})

      ? "aFields:"
      aeval (self:aRelPos, {|x| qout(x)})
      if !lPrintOnly
         wait "---ready, press any key: "
      endif
   endif
// #endif
   if lPrintOnly
      SET PRINTER (isprint)
      SET CONSOLE (iscons )
   endif
#endif
   return self:lUsed

*-----------------------------------------------------------------------------
* dummy, unused access methods here, 
* required only to display SELF in GUI debugger
*
ACCESS AliasSym      class smallRDD
   return NIL
ACCESS Concurrenc    class smallRDD
   return NIL
ACCESS Driver        class smallRDD
   return "SMALLRDD"
ACCESS ErrInfo       class smallRDD
   return NIL
ACCESS FileSpec      class smallRDD
   return NIL
ACCESS Filter        class smallRDD
   return NIL
ACCESS ForBlock      class smallRDD
   return NIL
ACCESS Found         class smallRDD
   return NIL
ACCESS GetLocate     class smallRDD
   return NIL
ACCESS Header        class smallRDD
   return NIL
ACCESS IndexCheck    class smallRDD
   return 0
ACCESS IndexExt      class smallRDD
   return NIL
ACCESS IndexKey      class smallRDD
   return NIL
ACCESS IndexLock     class smallRDD
   return NIL
ACCESS IsRelation    class smallRDD
   return NIL
ACCESS Lupdate       class smallRDD
   return .F.
ACCESS Name          class smallRDD
   return self:cName
ACCESS OrderBottom   class smallRDD
   return .F.
ACCESS OrderKeyNo    class smallRDD
   return 0
ACCESS OrderKeyVal   class smallRDD
   return NIL
ACCESS OrderTopSc    class smallRDD
   return NIL
ACCESS RddName       class smallRDD
   return "SMALLRDD"
ACCESS ReadOnly      class smallRDD
   return .F.
ACCESS RLockList     class smallRDD
   return NIL
ACCESS Scope         class smallRDD
   return NIL
ACCESS Status        class smallRDD
   return NIL
ACCESS WhileBlock    class smallRDD
   return NIL
   
*** eof smallRDD.prg ***

